package Util;

/** Utilities for the talking heads simulation.
 * This class contains all kinds of functions that appeared useful in various places. 
 * 
 * <p>
 * Copyright (c) 2004 Paul Vogt
 *
 * @author Paul Vogt
 * @version $4.0.3$

*/

import java.util.*;
import LGames.*;
import java.io.PrintWriter;
import javax.swing.text.DefaultEditorKit.BeepAction;
import javax.swing.plaf.metal.MetalLookAndFeel;

public class Utils{

    /**
     * Calculates the Eucledian distance between two arrays.
     *
     * @param x double array 1
     * @param y double array 2
     * @return distance between x and y, or -1 if the dimensions of x and y differ.
     */
    public static double distance(double [] x,double [] y){
	double sum=0;
	if (x.length==y.length)
	    for (int i=0;i<x.length;i++)
		sum+=(x[i]-y[i])*(x[i]-y[i]);
	else
	    sum=-1.0;

	return Math.sqrt(sum);
    }
    /**
     * This function calculates distances between 2 arrays where the arrays can have wild-cards (values <0).
     *
     * @param x double array 1
     * @param y double array 2
     * @return Eucledian distance between x and y for all non-negative dimensions
     */
    public static double distance1(double [] x,double [] y){
	double sum=0;
	if (x.length==y.length){
	    for (int i=0;i<x.length;i++)
		if (x[i]>=0.0 && y[i]>=0.0)
		    sum+=(x[i]-y[i])*(x[i]-y[i]);
	}
	else
	    sum=-1.0;

	return sum;
    }
    /**
     * This function calculates the average values of 2 vectors(double arrays)
     *
     * @param x vector 1
     * @param y vector 2
     * @return a vector where each element is the average of the two input vectors
     */
    public static double [] averageVector(double [] x,double [] y){
	double [] z = new double[x.length];
	if (x.length==y.length){
	    for (int i=0;i<x.length;i++){
		if (x[i]>=0.0 && y[i]>=0.0)
		    z[i]=0.5*(x[i]+y[i]);
		else z[i]=-1.0;
	    }
	}
	return z;
    }
    /**
     * Function writes a double array to a string where each double only gets 5 places, cf. doubleString(x,5)
     *
     * @param x the double array
     * @return the string
     */
    public static String doubleArrayString(double [] x){
	String retval=new String("[");
	for (int i=0;i<x.length;i++){
	    retval=retval.concat(doubleString(x[i],5));
	    if (i<x.length-1)
		retval=retval.concat(",");
	}
	retval=retval.concat("]");
	return retval;
    }
    /**
     * Function forms a string of a double value with a given string length. E.g. doubleString(0.33333333,5)
     * returns "0.333"
     *
     * @param x the double value
     * @param l the string length
     * @return string value of the double x with length l
     */
   public static String doubleString(final double x,final int l){
	String X=new String(String.valueOf(x));
	if (x>0.0 && x<1.0/Math.pow(10.0,l-2)){
	    X=new String("0.");
	    for (int i=3;i<=l;i++)
		X=X.concat("0");
	}
	else if (x<0.0 && x>1.0/Math.pow(10.0,l-2)){
	    X=new String("-0.");
	    for (int i=3;i<=l;i++)
		X=X.concat("0");
	}
	int le=Math.min(X.length(),l);
	return X.substring(0,le);
    }

    /**
     * Function to print the 'global' lexicon
     */
    public static void printGlobalLexicon(List lAgents,
					  PrintWriter lexFile,char type,
					  int iter){
	List bagOfWords = new ArrayList();
	String word;
	String meaning;
	lexFile.println("Global lexicon, iteration "+iter+":");
	lexFile.println();
	for (int i=0;i<lAgents.size();i++)
	    ((Agent)lAgents.get(i)).getWords(bagOfWords);
	for (int i=0;i<bagOfWords.size();i++){
	    word=(String)bagOfWords.get(i);
	    for (int j=0;j<lAgents.size();j++){
		meaning=((Agent)lAgents.get(j)).getMeaning(word,type);
		lexFile.println("A"+((Agent)lAgents.get(j)).getID()+" "+
				word+" "+meaning);
	    }
	}
	lexFile.println();
    }
    /**
     * Function to print the grammars of the entire population.
     */
    public static void printGrammar(List lAgents,
				    PrintWriter lexFile,char type,
				    int iter){
	lexFile.println("Grammar, iteration "+iter+":");
	lexFile.println();
	for (int i=0;i<lAgents.size();i++)
	    ((CompositionalAgent2)lAgents.get(i)).printGrammar(lexFile,type);
	lexFile.println();
	lexFile.flush();
    }
    /**
     * Function used at some point to output to a latex file
     */
    public static void beginLatexDocument(PrintWriter file,final String title){

	file.println("\\documentclass{article}");
	file.println("\\usepackage{mystyle}");
	file.println("\\title{"+title+"}");
	file.println("\\begin{document}");
	file.println("\\maketitle");
	file.println("\\tiny");
    }

    /**
     * Function used at some point to output to a latex file
     */
    public static void endLatexDocument(PrintWriter file){

	file.println("\\end{document}");
    }
    /**
     * Function used at some point to output to a latex file
     */

    public static void printLexiconLatexTabel(List lAgents,
					  PrintWriter lexFile,char type,int lg){

	for (int i=0;i<lAgents.size();i++)
	    ((Agent)lAgents.get(i)).printLexiconLatexTabel(lexFile,type,lg);

    }
    /**
     * Function to write the lexicon in some format to the lexFile
     */
    public static void printLexiconColumns(List lAgents,
					  PrintWriter lexFile,char type,int lg){
	    lexFile.print(lg);
	    for (int i=0;i<lAgents.size();i++)
		((Agent)lAgents.get(i)).printLexiconColumns(lexFile,type,lg);
	    lexFile.println();
    }
    /**
     * Function to remove a string from another string
     *
     * @param s1 string 1
     * @param s2 string 2
     * @return the remainder of string s1 after removal of string s2
     */
    public static String removeSubstring(final String s1,final String s2){
	int n=s1.indexOf(s2);
	int l=s2.length();
	String r=new String();
	if (n>=0){
	    r=r.concat(s1.substring(0,n));
	    if (l!=s1.length())
		r=r.concat(s1.substring(n+l,s1.length()));
	}
	else r=s1;
	return r;
    }
    /**
     * This function is used to get the largest substring under certain conditions. These conditions are:
     * if the substring starts or ends in one string, it must also start or end in the other, and it must
     * either start or end at the strings
     */
    public static String substring(final String s1,final String s2){
	String sub=new String();
	boolean eq=true;
	int n=0,i=0,j=0;
	if (s1.indexOf(s2)>=0 || s2.indexOf(s1)>=0) return sub;
	while (i<s1.length()){
	    while (j<s2.length()){
		if (i<s1.length() && s1.charAt(i)==s2.charAt(j)){
		    eq=true;
		    n=0;
		    while (eq){
			if (i+n<s1.length() && j+n<s2.length() && s1.charAt(i+n)==s2.charAt(j+n))
			    n++;
			else{ 
			    if (((i+n==s1.length() && j+n==s2.length()) ||//may be at end of string
				(i+n<s1.length() && j+n<s2.length() && 
				 s1.charAt(i+n)!=s2.charAt(j+n))) && n>1 &&//substring ended & l>1
				!((i==0 && j>0) || (i>0 && j==0)) &&//not allow sub at beginning 
				//of one string while not at the other
				(s1.substring(i,i+n)).length()>sub.length())//return largest substr
				sub=s1.substring(i,i+n);
			    eq=false;
			}
		    }
		}
		j++;
	    }
	    i++;j=1;
	}
	return sub;
    }
    /*
    private static double realSplit(final String s1,final String s2,List segments){
	//List segments=new ArrayList();
	double retval=0.0;
	int i=s2.indexOf(s1);
	if (s2.startsWith(s1)){
	    segments.add(s1);
	    segments.add(s2.substring(s1.length()));
	    retval=(double)(s1.length()*(s2.length()-s1.length()))/
		(double)(s2.length()*s2.length());
	}
	else if (s2.endsWith(s1)){
	    segments.add(s2.substring(0,i));
	    segments.add(s1);
	    retval=(double)(i*s1.length())/
		(double)(s2.length()*s2.length());
	}
	return retval;
    }

    public static double split(final String s1,final String s2,List segments1,List segments2){

	String s=new String();
	boolean finished=false;
	String sub=new String();

	sub=substring(s1,s2);
	double sim=0.0;

	if (sub!=""){
	    sim+=realSplit(sub,s1,segments1);
	    sim+=realSplit(sub,s2,segments2);

	}

	return sim;	
    }
    */
    /**
     * Function returns the complement of a string
     *
     * @param s1 string 1
     * @param s2 string 2
     * @return the substring of s1 that is unequal to s2, with the restriction that s1 must start or end with s2.
     * Otherwise it returns null.
     */
     public static String complement(final String s1,final String s2){
	String sub=null;
	if (s1.startsWith(s2))
	    sub=s1.substring(s2.length());
	else if (s1.endsWith(s2)) sub=s1.substring(0,s1.lastIndexOf(s2));
	return sub;
    }
    /**
     * Returns the largest substring of two strings if the substring is either at the start or at the end of
     * both strings and if the substring is not equal to either of the  two strings.
     *
     */
    public static String largestSubString(final String s1,final String s2){

	int maxL=0;

	//maxLength=Math.min(s1.length,s2.length)-1;

	if (s1.equals(s2)) return new String();//both strings should not be equal
	if (s1.startsWith(s2) || s1.endsWith(s2) || s2.startsWith(s1) || s2.endsWith(s1)) return new String();
	String sub = new String();
	for (int i=1;i<s2.length();i++)
	    if (s1.startsWith(s2.substring(0,i))) maxL=i;
	for (int i=1;i<s2.length()-1;i++){
	    sub=s2.substring(i);
	    if (s1.endsWith(sub) && sub.length()>maxL)
		return new String(sub);
	}
	return s2.substring(0,maxL);
    }
    /**
     * @param l a list of IntArrays
     * @param a an IntArray
     * @return the index of a in l (-1 if a is not in l)
     */
    public static int indexOf(final List l,final IntArray a){
	for (int i=0;i<l.size();i++)
	    if (((IntArray)l.get(i)).containsAll(a)) return i;
	return -1;
    }

    public static int indexOfPart(final List l,final IntArray a){
	for (int i=0;i<l.size();i++)
	    if (((IntArray)l.get(i)).containsElementOfEqualSize(a)) return i;
	return -1;
    }

    public static int indexOfContainingElement(final List l,final IntArray a){
	for (int i=0;i<l.size();i++)
	    if (((IntArray)l.get(i)).containsElementOf(a)) return i;
	return -1;
    }
    /**
     * Error message. Outputs the error message to the standard output and exits the program.
     */
    public static void error(final String mess){
	System.out.println("Error message: "+mess);
	System.exit(1);
    }
    /**
     * Function to calculate the average variance in the columns of a matrix. This is used to 
     * calculate the variance between a set of meanings.
     *
     * @param x the matrix
     * @param N the number of columns
     * @param dim the number of rows
     * @return the avarage variance of the columns
     */
    public static double variance(final double [][] x,final int N,final int dim){
	//to calculate the average population variance of the columns of x
	double sum,var=0.0,avg;
	int n;
	for (int i=0;i<dim;i++){
	    sum=0.0;
	    n=0;
	    for (int j=0;j<N;j++) 
		if (x[j][i]>=0.0){ 
		    sum+=x[j][i];n++;
		}
	    avg=sum/(double)n;
	    sum=0.0;
	    for (int j=0;j<N;j++) 
		if (x[j][i]>=0){ 
		    sum+=(x[j][i]-avg)*(x[j][i]-avg);
		}
	    if (n>0)
		var+=sum/(double)n;
	}
	return var/(double)dim;
    }
    /**
     * @return a string of an array of strings
     */
    public static String printStringArray(final String [] s){
	String retval=new String("[");
	for (int i=0;i<s.length;i++){
	    retval=retval.concat(s[i]);
	    if (i<s.length-1) retval=retval.concat(",");
	}
	retval=retval.concat("]");
	return retval;
    }
    /**
     * @param x a double value (between 0 and 1)
     * @return an integer between 0 and 9: 0 if x<0.1, 1 if 0.1<=x<0.2, ... ,9 if 0.9<=x
     */
    public static int index(double x){
	if (x<0.1) return 0;
	else if (x<0.2) return 1;
	else if (x<0.3) return 2;
	else if (x<0.4) return 3;
	else if (x<0.5) return 4;
	else if (x<0.6) return 5;
	else if (x<0.7) return 6;
	else if (x<0.8) return 7;
	else if (x<0.9) return 8;
	return 9;
    }


}
